# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for recsim.agents.layers.fixed_length_history."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from gym import spaces
import mock
from recsim.agents import cluster_bandit_agent
from recsim.agents.layers import fixed_length_history
import tensorflow.compat.v1 as tf


class FixedLengthHistoryTest(tf.test.TestCase):

  def setUp(self):
    self.history_length = 3
    self.slate_size = 1
    super(FixedLengthHistoryTest, self).setUp()
    self.test_action_space = mock.Mock(nvec=mock.Mock(shape=[2,]))
    self.test_observation_space = spaces.Dict({
        'user': spaces.Discrete(2),
        'response': spaces.Discrete(3),
        'doc': spaces.Tuple((spaces.Discrete(4),))
    })
    self.mock_agent = mock.create_autospec(
        cluster_bandit_agent.ClusterBanditAgent)
    self.history = fixed_length_history.FixedLengthHistoryLayer(
        self.mock_agent,
        self.test_observation_space,
        self.test_action_space,
        self.history_length,
        remember_user=True,
        remember_response=True,
        remember_doc=True,
        kwarg_for_agent=-1)

  def test_initialization(self):
    # Basic object properties initialized correctly.
    self.assertEqual(self.history._history_length, self.history_length)
    self.assertCountEqual(self.history._features, ['user', 'doc', 'response'])
    # Base agent got constructed correctly.
    self.mock_agent.assert_called_once()
    mock_agent_args, mock_agent_kwargs = self.mock_agent.call_args
    self.assertEmpty(mock_agent_args)
    self.assertCountEqual(
        mock_agent_kwargs.keys(),
        ['observation_space', 'action_space', 'kwarg_for_agent'])
    # Environment features and kwargs are passed down to base agent.
    self.assertEqual(mock_agent_kwargs['action_space'], self.test_action_space)
    self.assertEqual(mock_agent_kwargs['kwarg_for_agent'], -1)
    # Augmented observation is space properly formatted.
    augmented_observation_space = mock_agent_kwargs['observation_space']
    for field in ['doc', 'response']:
      self.assertEqual(augmented_observation_space.spaces[field],
                       self.test_observation_space[field])
    self.assertCountEqual(
        augmented_observation_space.spaces['user'].spaces.keys(),
        ['raw_observation', 'sufficient_statistics'])
    self.assertEqual(
        augmented_observation_space.spaces['user']['raw_observation'],
        self.test_observation_space['user'])
    self.assertLen(
        augmented_observation_space.spaces['user']
        .spaces['sufficient_statistics'].spaces, self.history_length)
    for space in (augmented_observation_space.spaces['user']
                  .spaces['sufficient_statistics'].spaces):
      self.assertEqual(space, self.test_observation_space)

  def test_update_and_observation(self):
    observation = self.test_observation_space.sample()
    self.assertIsNone(self.history._sufficient_statistics)
    self.history._update(observation)
    self.assertEqual(self.history._sufficient_statistics,
                     [observation, None, None])
    self.assertEqual(self.history._create_observation(),
                     (observation, None, None))

if __name__ == '__main__':
  tf.test.main()



# coding=utf-8
# Lint as: python3
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for recsim.agent."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import functools
import os

from absl.testing import parameterized
import gym
import numpy as np
from recsim import agent
from recsim.simulator import environment
from recsim.simulator import recsim_gym
from recsim.simulator import runner_lib
import tensorflow.compat.v1 as tf


class DummySingleUserAgent(agent.AbstractEpisodicRecommenderAgent):

  def step(self, reward, observation):
    pass

  def bundle_and_checkpoint(self, checkpoint_dir, iteration_number):
    pass

  def unbundle(self, checkpoint_dir, iteration_number, bundle_dict):
    pass


class DummyMultiUserAgent(agent.AbstractMultiUserEpisodicRecommenderAgent):

  def step(self, reward, observation):
    pass

  def bundle_and_checkpoint(self, checkpoint_dir, iteration_number):
    pass

  def unbundle(self, checkpoint_dir, iteration_number, bundle_dict):
    pass


class DummySingleUserEnvironment(environment.SingleUserEnvironment):

  def __init__(self):
    pass


class DummyMultiUserEnvironment(environment.MultiUserEnvironment):

  def __init__(self):
    pass


def create_environment(env_config):
  if env_config['multiuser_env']:
    env = DummyMultiUserEnvironment()
  else:
    env = DummySingleUserEnvironment()
  reward_aggregator = lambda x: x
  return recsim_gym.RecSimGymEnv(env, reward_aggregator)


class AgentTest(parameterized.TestCase):
  num_users = 100
  num_candidates = 10
  slate_size = 8

  @parameterized.named_parameters(
      ('multiuser_env_multiuser_agent', True, True, True),
      ('singleuser_env_multiuser_agent', False, True, False),
      ('multiuser_env_singleuser_agent', True, False, False),
      ('singleuser_env_singleuser_agent', False, False, True),
  )
  def test_multi_and_single_user_consistency(
      self, multiuser_env, multiuser_agent, should_succeed):

    def create_agent(
        sess, env, summary_writer, eval_mode, multiuser_agent=True):
      del sess, env, summary_writer, eval_mode  # unused
      action_space = gym.spaces.MultiDiscrete(np.ones((self.slate_size,)))
      if multiuser_agent:
        action_space = gym.spaces.Tuple([action_space] * self.num_users)
        AgentClass = DummyMultiUserAgent  # pylint:disable=invalid-name
      else:
        AgentClass = DummySingleUserAgent  # pylint:disable=invalid-name
      return AgentClass(action_space)

    env_config = dict(multiuser_env=multiuser_env)
    base_dir = '/tmp/Env%sAgent%s' % (multiuser_env, multiuser_agent)
    create_agent_fn = functools.partial(create_agent,
                                        multiuser_agent=multiuser_agent)
    if not os.path.exists(base_dir):
      os.makedirs(base_dir)
    if should_succeed:  # constructors should work
      _ = runner_lib.TrainRunner(
          base_dir=base_dir,
          create_agent_fn=create_agent_fn,
          env=create_environment(env_config),
          max_training_steps=1,
          max_steps_per_episode=1,
          num_iterations=1)
    else:  # agent constructor should raise error
      with self.assertRaises(ValueError):
        _ = runner_lib.TrainRunner(
            base_dir=base_dir,
            create_agent_fn=create_agent_fn,
            env=create_environment(env_config),
            max_training_steps=1,
            max_steps_per_episode=1,
            num_iterations=1)


if __name__ == '__main__':
  tf.test.main()



# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for recsim.agents.bandits.glm_algorithm."""

import numpy as np
from recsim.agents.bandits import glm_algorithms
from scipy import special
import tensorflow.compat.v1 as tf


class UCB_GLMTest(tf.test.TestCase):  # pylint: disable=invalid-name

  def setUp(self):
    super(UCB_GLMTest, self).setUp()
    self._dim = 3
    self._alg = glm_algorithms.UCB_GLM(self._dim, horizon=100)

  def add_random_arms(self, n_arms):
    # Add some random arms with rewards
    for _ in range(n_arms):
      arm = np.random.uniform(size=self._alg._dim)
      reward = np.random.binomial(1, 0.5)  # Random reward
      self._alg.update(reward, arm)

  def test_update(self):
    n_arms = 10
    self.add_random_arms(n_arms)
    self.assertLen(self._alg._arms, n_arms)

  def test_arm_matrix(self):
    n_arms = 10
    arms = [np.random.uniform(size=self._dim) for _ in range(n_arms)]
    arm_matrix = self._alg.get_arm_matrix(arms)
    self.assertEqual(np.shape(arm_matrix), (n_arms, self._alg._dim))

  def test_solve_logistic_bandit(self):
    n_arms = 10
    self.add_random_arms(n_arms)
    w, gram = self._alg.solve_logistic_bandit()
    self.assertEqual(np.shape(w), (self._dim,))
    self.assertEqual(np.shape(gram), (self._dim, self._dim))

  def test_learning(self):
    # Construct a set of arms
    n_arms = 10
    arms = [np.random.uniform(size=self._dim) for _ in range(n_arms)]
    # Sample weight vector
    w_star = np.random.normal(size=self._dim)
    rounds = 20
    for _ in range(rounds):
      arm, arm_id, scores = self._alg.get_arm(arms)
      self.assertLess(arm_id, len(arms))
      self.assertLen(scores, len(arms))
      reward = np.random.binomial(1, special.expit(np.dot(arm, w_star)))
      self._alg.update(reward, arm)


class GLM_TSTest(tf.test.TestCase):  # pylint: disable=invalid-name

  def setUp(self):
    super(GLM_TSTest, self).setUp()
    self._dim = 3
    self._alg = glm_algorithms.GLM_TS(self._dim)

  def test_learning(self):
    # Construct a set of arms
    n_arms = 20
    arms = [np.random.uniform(size=self._dim) for _ in range(n_arms)]
    # Sample weight vector
    w_star = np.random.normal(size=self._dim)
    rounds = 10
    for _ in range(rounds):
      arm, arm_id, scores = self._alg.get_arm(arms)
      self.assertLess(arm_id, len(arms))
      self.assertLen(scores, len(arms))
      reward = np.random.binomial(1, special.expit(np.dot(arm, w_star)))
      self._alg.update(reward, arm)


if __name__ == '__main__':
  tf.test.main()



# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Classes for Bandit Algorithms for Generalized Linear Models."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import abc
import numpy as np
from scipy import special
import six


@six.add_metaclass(abc.ABCMeta)
class GLMAlgorithm(object):
  """Base class for Generalized Linear Models (GLM) bandit algorithms.

  In this setting each arm is represented by a feature vector x and there exists
  an unknown weight vector w*. In each round the algorithm pulls an arm and gets
  a noisy reward, which is assumed to be the result of composing the dot product
  x.w* with a link function, plus random noise. For example,
  sigmoid(x.w*) + eps, where eps is sub-Gaussian noise.

  Attributes:
    arms: the arms pulled so far
    rewards: the rewards observed so far
    dim: the dimension of the feature/weight vectors
    outer: the sum of outer products of the arms pulled so far (x.x^T)
    sigma0: a parameter scaling the identity matrix used for making the
      resulting Gram matrix positive definite.
    optimism_scaling: A float specifying the confidence level. Default value
      (1.0) corresponds to the exploration strategy presented in the literature.
      A smaller number means less exploration and more exploitation.
    _rng: An instance of random.RandomState for random number generation
  """

  def __init__(self, dim, sigma0=1., optimism_scaling=1.):
    self._rewards = np.array([])  # stores all rewards
    self._arms = np.ndarray([0, dim])  # stores arm differences
    self._dim = dim
    self._outer = np.zeros([dim, dim])
    self._sigma0 = sigma0
    self._optimism_scaling = optimism_scaling

  def update(self, reward, arm):
    """Updates state with arm and reward.

    Args:
      reward: the reward received
      arm: the arm that was pulled
    """
    assert len(arm) == self._dim, 'Expected dimension {}, got {}'.format(
        self._dim, len(arm))
    self._rewards = np.append(self._rewards, reward)
    self._arms = np.concatenate((self._arms, [arm]), axis=0)
    self._outer += np.outer(arm, arm)

  def solve_logistic_bandit(self, init_iters=10, num_iters=20, tol=1e-3):
    """Solves the maximum-likelihood problem.

    Implements iterative reweighted least squares for Bayesian logistic
    regression. See sections 4.3.3 and 4.5.1 in Pattern Recognition and Machine
    Learning, Bishop (2006)

    Args:
      init_iters: number of initial iterations to skip (returns zeros)
      num_iters: number of least squares iterations
      tol: tolerance level of change in solution between iterations before
        terminating
    Returns:
      w: maximum likelihood solution
      gram: Gram matrix
    """

    arms = self._arms
    w = np.zeros(self._dim)
    gram = np.eye(self._dim) / np.square(self._sigma0)
    if len(self._arms) > init_iters:
      for _ in range(num_iters):
        prev_w = np.copy(w)
        arms_w = arms.dot(w)
        sig_arms_w = special.expit(arms_w)
        r = np.diag(sig_arms_w * (1 - sig_arms_w))
        gram = (((arms.T).dot(r)).dot(arms) +
                np.eye(self._dim) / np.square(self._sigma0))
        rz = r.dot(arms_w) - (sig_arms_w - self._rewards)
        w = np.linalg.solve(gram, (arms.T).dot(rz))
        if np.linalg.norm(w - prev_w) < tol:
          break

    return w, gram

  def get_arm_matrix(self, arms):
    """Puts all arms into a matrix."""
    return np.stack(arms, axis=0)

  @abc.abstractmethod
  def get_arm(self, arms):
    """Computes which arm to pull next.

    Args:
      arms: a list of feature vectors, one for each arm

    Returns:
      arm: the chosen arm
      arm_ind: index of the chosen arm
      scores: an array with arm scores
    """


class UCB_GLM(GLMAlgorithm):  # pylint: disable=invalid-name
  """UCB-GLM algorithm.

  See "Provably Optimal Algorithms for Generalized Linear Contextual Bandits",
  by Li et al. (2017).
  """

  def __init__(self, dim, horizon, sigma0=1., optimism_scaling=1.):
    super(UCB_GLM, self).__init__(dim, sigma0, optimism_scaling)
    # Set confidence interval scaling, by
    # Theorem 2 in Li (2017)
    # Provably Optimal Algorithms for Generalized Linear Contextual Bandits
    crs = optimism_scaling  # confidence region scaling
    delta = 1. / float(horizon)
    sigma = 0.5
    kappa = 0.25
    # Confidence ellipsoid width (cew):
    cew = (sigma / kappa) * (np.sqrt((self._dim / 2) *
                                     np.log(1. + 2. * horizon / self._dim) +
                                     np.log(1 / delta)))
    self._ci_scaling = crs * cew

  def get_arm(self, arms):
    """Computes which arm to pull next.

    Args:
      arms: a list of feature vectors, one for each arm
    Returns:
      The selected arm, its index in arms, and the computed scores
    """
    arm_matrix = self.get_arm_matrix(arms)
    gram = self._outer + np.eye(self._dim) / np.square(self._sigma0)
    gram_inv = np.linalg.inv(gram)
    ucbs = np.sqrt((np.matmul(arm_matrix, gram_inv) * arm_matrix).sum(axis=1))
    # Estimate w
    w, _ = self.solve_logistic_bandit()
    # Compute UCB
    mu = np.matmul(arm_matrix, w) + self._ci_scaling * ucbs
    arm = np.random.choice(np.flatnonzero(mu == mu.max()))

    return arms[arm], arm, mu

  @staticmethod
  def print():
    return 'GLM-UCB'


class GLM_TS(GLMAlgorithm):  # pylint: disable=invalid-name
  """Thompson sampling algorithm for generalized linear models.

  See "Linear Thompson Sampling Revisited" by Abeille and Lazaric (2017).
  """

  def get_arm(self, arms):
    """Computes which arm to pull next.

    Args:
      arms: a list of feature vectors, one for each arm
    Returns:
      The selected arm, its index in arms, and the computed scores
    """
    arm_matrix = self.get_arm_matrix(arms)
    w, gram = self.solve_logistic_bandit()
    gram_inv = np.square(self._optimism_scaling) * np.linalg.inv(gram)

    # Posterior sampling
    w_tilde = np.random.multivariate_normal(w, gram_inv)
    mu = np.matmul(arm_matrix, w_tilde)
    # Argmax breaking ties randomly
    arm = np.random.choice(np.flatnonzero(mu == mu.max()))

    return arms[arm], arm, mu

  @staticmethod
  def print():
    return 'GLM-TS'



# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Convenience primitives relating to the implementation of agents."""
from gym import spaces
import numpy as np


class GymSpaceWalker(object):
  """Class for recursively applying a given function to a gym space.

  Gym spaces have nested structure in terms of container spaces (e.g. Dict and
  Tuple) containing basic spaces such as Discrete and Box. This class consumes a
  gym observation definition and a leaf operator is used to produce a flat list
  of the contents of the gym space, apply the leaf operator to all basic spaces
  in the proces. E.g., given a gym space of the form Tuple((Box(1), Box(1)) and
  a leaf operator f, this class can is used to transform an observation (a, b)
  to [f(a), f(b)].

  Args:
  gym_space: An instance of an OpenAI Gym space.
  leaf_op: A function taking as arguments an OpenAI Gym space and an observation
    conforming to that space. There are no requirements on its output.
  """

  def __init__(self, gym_space, leaf_op):
    self._gym_space = gym_space
    self._leaf_op = leaf_op

  def apply_and_flatten(self, gym_observations):
    return self._descend_and_flatten(self._gym_space, gym_observations)

  def _descend_and_flatten(self, gym_space, gym_observations):
    """Recursive implementation of flattening and leaf op application.

    Args:
      gym_space: An instance of an OpenAI Gym space.
      gym_observations: A list of observation conforming to the format
        of gym_space.

    Returns:
      flattened_apply: a list of the applications of leaf_op to the leaves of
        the gym space, as encountered in post-order traversal.
    """
    if isinstance(gym_space, spaces.dict.Dict):
      flattened_apply = []
      for key, space in gym_space.spaces.items():
        flattened_apply += self._descend_and_flatten(
            space,
            [gym_observation[key] for gym_observation in gym_observations])
    elif isinstance(gym_space, spaces.tuple.Tuple):
      flattened_apply = []
      for i, space in enumerate(gym_space.spaces):
        flattened_apply += self._descend_and_flatten(
            space, [gym_observation[i] for gym_observation in gym_observations])
    elif isinstance(gym_space, spaces.box.Box) or isinstance(
        gym_space, spaces.discrete.Discrete):
      return self._leaf_op(gym_space, gym_observations)
    else:
      raise NotImplementedError('Gym space type ' + str(type(gym_space)) +
                                ' not implemented yet.')
    return flattened_apply


def epsilon_greedy_exploration(state_action_iterator, q_function, epsilon):
  """Epsilon greedy exploration.

  Either picks a slate uniformly at random with probability epsilon, or returns
  a slate with maximal Q-value. TODO(mmladenov): more verbose doc.
  Args:
    state_action_iterator: an iterator over slate, state_action_index tuples.
    q_function: a container holding Q-values of state-action pairs.
    epsilon: probability of random action.
  Returns:
    slate: the picked slate.
    sa_index: the index of the picked slate in the Q-value table.
  """
  max_q_next = -np.Inf
  max_state_action_index = None
  max_slate = []
  random_state_action_index = None
  random_slate = []
  sa_index = None
  for slate_count, (slate, state_action_index) in enumerate(
      state_action_iterator, start=1):
    q_value = q_function.get(state_action_index, 0)
    if q_value > max_q_next:
      max_q_next = q_value
      max_state_action_index = state_action_index
      max_slate = slate
    # Pick a random action by reservoir sampling in order to avoid materializing
    # all possible slates.
    if slate_count == 1 or np.random.random() < 1.0 / (1.0 * slate_count):
      random_state_action_index = state_action_index
      random_slate = slate
  if np.random.random() <= epsilon:
    slate = random_slate
    sa_index = random_state_action_index
  else:
    slate = max_slate
    sa_index = max_state_action_index
  return slate, sa_index


def min_count_exploration(state_action_iterator, counts_function):
  """Minimum count exploration.

  Picks the state-action pair with minimum counts.
  Args:
    state_action_iterator: an iterator over slate, state_action_index tuples.
    counts_function: a container holding the number of times a state-action pair
      has been executed so far.
  Returns:
    slate: the picked slate.
    sa_index: the index of the picked slate in the counts table.
  """
  min_sa_count = np.Inf
  min_sa_count_slate = []
  min_sa_count_index = None
  for slate, state_action_index in state_action_iterator:
    sa_count = counts_function.get(state_action_index, 0)
    if sa_count < min_sa_count:
      min_sa_count_slate = slate
      min_sa_count_index = state_action_index
      min_sa_count = sa_count
    if sa_count == 0:
      break
  return min_sa_count_slate, min_sa_count_index



# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Agent that implements the Slate-Q algorithms."""
import gin.tf
import numpy as np
from recsim import agent as abstract_agent
from recsim import choice_model
from recsim.agents.dopamine import dqn_agent
import tensorflow.compat.v1 as tf


def compute_probs_tf(slate, scores_tf, score_no_click_tf):
  """Computes the selection probability and returns selected index.

  This assumes scores are normalizable, e.g., scores cannot be negative.

  Args:
    slate: a list of integers that represents the video slate.
    scores_tf: a float tensor that stores the scores of all documents.
    score_no_click_tf: a float tensor that represents the score for the action
      of picking no document.

  Returns:
    A float tensor that represents the probabilities of selecting each document
      in the slate.
  """
  all_scores = tf.concat([
      tf.gather(scores_tf, slate),
      tf.reshape(score_no_click_tf, (1, 1))
  ], axis=0)  # pyformat: disable
  all_probs = all_scores / tf.reduce_sum(input_tensor=all_scores)
  return all_probs[:-1]


def score_documents_tf(user_obs,
                       doc_obs,
                       no_click_mass=1.0,
                       is_mnl=False,
                       min_normalizer=-1.0):
  """Computes unnormalized scores given both user and document observations.

  This implements both multinomial proportional model and multinormial logit
    model given some parameters. We also assume scores are based on inner
    products of user_obs and doc_obs.

  Args:
    user_obs: An instance of AbstractUserState.
    doc_obs: A numpy array that represents the observation of all documents in
      the candidate set.
    no_click_mass: a float indicating the mass given to a no click option
    is_mnl: whether to use a multinomial logit model instead of a multinomial
      proportional model.
    min_normalizer: A float (<= 0) used to offset the scores to be positive when
      using multinomial proportional model.

  Returns:
    A float tensor that stores unnormalzied scores of documents and a float
      tensor that represents the score for the action of picking no document.
  """
  user_obs = tf.reshape(user_obs, [1, -1])
  scores = tf.reduce_sum(input_tensor=tf.multiply(user_obs, doc_obs), axis=1)
  all_scores = tf.concat([scores, tf.constant([no_click_mass])], axis=0)
  if is_mnl:
    all_scores = tf.nn.softmax(all_scores)
  else:
    all_scores = all_scores - min_normalizer
  return all_scores[:-1], all_scores[-1]


def score_documents(user_obs,
                    doc_obs,
                    no_click_mass=1.0,
                    is_mnl=False,
                    min_normalizer=-1.0):
  """Computes unnormalized scores given both user and document observations.

  Similar to score_documents_tf but works on NumPy objects.

  Args:
    user_obs: An instance of AbstractUserState.
    doc_obs: A numpy array that represents the observation of all documents in
      the candidate set.
    no_click_mass: a float indicating the mass given to a no click option
    is_mnl: whether to use a multinomial logit model instead of a multinomial
      proportional model.
    min_normalizer: A float (<= 0) used to offset the scores to be positive when
      using multinomial proportional model.

  Returns:
    A float array that stores unnormalzied scores of documents and a float
      number that represents the score for the action of picking no document.
  """
  scores = np.array([])
  for doc in doc_obs:
    scores = np.append(scores, np.dot(user_obs, doc))

  all_scores = np.append(scores, no_click_mass)
  if is_mnl:
    all_scores = choice_model.softmax(all_scores)
  else:
    all_scores = all_scores - min_normalizer
  assert not all_scores[
      all_scores < 0.0], 'Normalized scores have non-positive elements.'
  return all_scores[:-1], all_scores[-1]


def select_slate_topk(slate_size, s_no_click, s, q):
  """Selects the slate using the top-K algorithm.

  This algorithm corresponds to the method "TS" in
  Ie et al. https://arxiv.org/abs/1905.12767.

  Args:
    slate_size: int, the size of the recommendation slate.
    s_no_click: float tensor, the score for not clicking any document.
    s: [num_of_documents] tensor, the scores for clicking documents.
    q: [num_of_documents] tensor, the predicted q values for documents.

  Returns:
    [slate_size] tensor, the selected slate.
  """
  del s_no_click  # Unused argument.
  _, output_slate = tf.math.top_k(s * q, k=slate_size)
  return output_slate


def select_slate_greedy(slate_size, s_no_click, s, q):
  """Selects the slate using the adaptive greedy algorithm.

  This algorithm corresponds to the method "GS" in
  Ie et al. https://arxiv.org/abs/1905.12767.

  Args:
    slate_size: int, the size of the recommendation slate.
    s_no_click: float tensor, the score for not clicking any document.
    s: [num_of_documents] tensor, the scores for clicking documents.
    q: [num_of_documents] tensor, the predicted q values for documents.

  Returns:
    [slate_size] tensor, the selected slate.
  """

  def argmax(v, mask):
    return tf.argmax(
        input=(v - tf.reduce_min(input_tensor=v) + 1) * mask, axis=0)

  numerator = tf.constant(0.)
  denominator = tf.constant(0.) + s_no_click
  mask = tf.ones(tf.shape(input=q)[0])

  def set_element(v, i, x):
    mask = tf.one_hot(i, tf.shape(input=v)[0])
    v_new = tf.ones_like(v) * x
    return tf.where(tf.equal(mask, 1), v_new, v)

  for _ in range(slate_size):
    k = argmax((numerator + s * q) / (denominator + s), mask)
    mask = set_element(mask, k, 0)
    numerator = numerator + tf.gather(s * q, k)
    denominator = denominator + tf.gather(s, k)

  output_slate = tf.where(tf.equal(mask, 0))
  return output_slate


def select_slate_optimal(slate_size, s_no_click, s, q):
  """Selects the slate using exhaustive search.

  This algorithm corresponds to the method "OS" in
  Ie et al. https://arxiv.org/abs/1905.12767.

  Args:
    slate_size: int, the size of the recommendation slate.
    s_no_click: float tensor, the score for not clicking any document.
    s: [num_of_documents] tensor, the scores for clicking documents.
    q: [num_of_documents] tensor, the predicted q values for documents.

  Returns:
    [slate_size] tensor, the selected slate.
  """

  num_candidates = s.shape.as_list()[0]

  # Obtain all possible slates given current docs in the candidate set.
  mesh_args = [list(range(num_candidates))] * slate_size
  slates = tf.stack(tf.meshgrid(*mesh_args), axis=-1)
  slates = tf.reshape(slates, shape=(-1, slate_size))

  # Filter slates that include duplicates to ensure each document is picked
  # at most once.
  unique_mask = tf.map_fn(
      lambda x: tf.equal(tf.size(input=x), tf.size(input=tf.unique(x)[0])),
      slates,
      dtype=tf.bool)
  slates = tf.boolean_mask(tensor=slates, mask=unique_mask)

  slate_q_values = tf.gather(s * q, slates)
  slate_scores = tf.gather(s, slates)
  slate_normalizer = tf.reduce_sum(
      input_tensor=slate_scores, axis=1) + s_no_click

  slate_q_values = slate_q_values / tf.expand_dims(slate_normalizer, 1)
  slate_sum_q_values = tf.reduce_sum(input_tensor=slate_q_values, axis=1)
  max_q_slate_index = tf.argmax(input=slate_sum_q_values)
  return tf.gather(slates, max_q_slate_index, axis=0)


def compute_target_sarsa(reward, gamma, next_actions, next_q_values,
                         next_states, terminals):
  """Computes the SARSA target Q value.

  Args:
    reward: [batch_size] tensor, the immediate reward.
    gamma: float, discount factor with the usual RL meaning.
    next_actions: [batch_size, slate_size] tensor, the next slate.
    next_q_values: [batch_size, num_of_documents] tensor, the q values of the
      documents in the next step.
    next_states: [batch_size, 1 + num_of_documents] tensor, the features for the
      user and the docuemnts in the next step.
    terminals: [batch_size] tensor, indicating if this is a terminal step.

  Returns:
    [batch_size] tensor, the target q values.
  """
  stack_number = -1
  user_obs = next_states[:, 0, :, stack_number]
  doc_obs = next_states[:, 1:, :, stack_number]

  batch_size = next_q_values.get_shape().as_list()[0]
  next_sarsa_q_list = []
  for i in range(batch_size):
    s, s_no_click = score_documents_tf(user_obs[i], doc_obs[i])
    q = next_q_values[i]

    slate = tf.expand_dims(next_actions[i], 1)
    p_selected = compute_probs_tf(slate, s, s_no_click)
    q_selected = tf.gather(q, slate)
    next_sarsa_q_list.append(
        tf.reduce_sum(input_tensor=p_selected * q_selected))

  next_sarsa_q_values = tf.stack(next_sarsa_q_list)

  return reward + gamma * next_sarsa_q_values * (1. -
                                                 tf.cast(terminals, tf.float32))


# The top-K and greedy algorithms rely on the fact that the
# probs[slate] = normalize(scores[slate], score_no_click)
def compute_target_greedy_q(reward, gamma, next_actions, next_q_values,
                            next_states, terminals):
  """Computes the optimal target Q value with the adaptive greedy algorithm.

  This algorithm corresponds to the method "GT" in
  Ie et al. https://arxiv.org/abs/1905.12767..

  Args:
    reward: [batch_size] tensor, the immediate reward.
    gamma: float, discount factor with the usual RL meaning.
    next_actions: [batch_size, slate_size] tensor, the next slate.
    next_q_values: [batch_size, num_of_documents] tensor, the q values of the
      documents in the next step.
    next_states: [batch_size, 1 + num_of_documents] tensor, the features for the
      user and the docuemnts in the next step.
    terminals: [batch_size] tensor, indicating if this is a terminal step.

  Returns:
    [batch_size] tensor, the target q values.
  """
  slate_size = next_actions.get_shape().as_list()[1]
  stack_number = -1
  user_obs = next_states[:, 0, :, stack_number]
  doc_obs = next_states[:, 1:, :, stack_number]

  batch_size = next_q_values.get_shape().as_list()[0]
  next_greedy_q_list = []
  for i in range(batch_size):
    s, s_no_click = score_documents_tf(user_obs[i], doc_obs[i])
    q = next_q_values[i]

    slate = select_slate_greedy(slate_size, s_no_click, s, q)
    p_selected = compute_probs_tf(slate, s, s_no_click)
    q_selected = tf.gather(q, slate)
    next_greedy_q_list.append(
        tf.reduce_sum(input_tensor=p_selected * q_selected))

  next_greedy_q_values = tf.stack(next_greedy_q_list)

  return reward + gamma * next_greedy_q_values * (
      1. - tf.cast(terminals, tf.float32))


def _get_unnormalized_scores(states):
  """Computes the unnormalized scores for the docs."""
  stack_number = -1
  user_obs = states[:, 0, :, stack_number]
  doc_obs = states[:, 1:, :, stack_number]

  batch_size = states.get_shape().as_list()[0]
  scores_list = []
  score_no_click_list = []
  for i in range(batch_size):
    scores_tf, score_no_click_tf = score_documents_tf(user_obs[i], doc_obs[i])
    scores_list.append(scores_tf)
    score_no_click_list.append(score_no_click_tf)
  scores = tf.stack(scores_list)
  score_no_click = tf.stack(score_no_click_list)

  return scores, score_no_click


def compute_target_topk_q(reward, gamma, next_actions, next_q_values,
                          next_states, terminals):
  """Computes the optimal target Q value with the greedy algorithm.

  This algorithm corresponds to the method "TT" in
  Ie et al. https://arxiv.org/abs/1905.12767.

  Args:
    reward: [batch_size] tensor, the immediate reward.
    gamma: float, discount factor with the usual RL meaning.
    next_actions: [batch_size, slate_size] tensor, the next slate.
    next_q_values: [batch_size, num_of_documents] tensor, the q values of the
      documents in the next step.
    next_states: [batch_size, 1 + num_of_documents] tensor, the features for the
      user and the docuemnts in the next step.
    terminals: [batch_size] tensor, indicating if this is a terminal step.

  Returns:
    [batch_size] tensor, the target q values.
  """
  slate_size = next_actions.get_shape().as_list()[1]
  scores, score_no_click = _get_unnormalized_scores(next_states)

  # Choose the documents with top affinity_scores * Q values to fill a slate and
  # treat it as if it is the optimal slate.
  unnormalized_next_q_target = next_q_values * scores
  _, topk_optimal_slate = tf.math.top_k(
      unnormalized_next_q_target, k=slate_size)

  # Get the expected Q-value of the slate containing top-K items.
  # [batch_size, slate_size]
  next_q_values_selected = tf.batch_gather(
      next_q_values, tf.cast(topk_optimal_slate, dtype=tf.int32))

  # Get normalized affinity scores on the slate.
  # [batch_size, slate_size]
  scores_selected = tf.batch_gather(scores,
                                    tf.cast(topk_optimal_slate, dtype=tf.int32))

  next_q_target_topk = tf.reduce_sum(
      input_tensor=next_q_values_selected * scores_selected, axis=1) / (
          tf.reduce_sum(input_tensor=scores_selected, axis=1) + score_no_click)

  return reward + gamma * next_q_target_topk * (
      1. - tf.cast(terminals, tf.float32))


def compute_target_optimal_q(reward, gamma, next_actions, next_q_values,
                             next_states, terminals):
  """Builds an op used as a target for the Q-value.

  This algorithm corresponds to the method "OT" in
  Ie et al. https://arxiv.org/abs/1905.12767..

  Args:
    reward: [batch_size] tensor, the immediate reward.
    gamma: float, discount factor with the usual RL meaning.
    next_actions: [batch_size, slate_size] tensor, the next slate.
    next_q_values: [batch_size, num_of_documents] tensor, the q values of the
      documents in the next step.
    next_states: [batch_size, 1 + num_of_documents] tensor, the features for the
      user and the docuemnts in the next step.
    terminals: [batch_size] tensor, indicating if this is a terminal step.

  Returns:
    [batch_size] tensor, the target q values.
  """
  scores, score_no_click = _get_unnormalized_scores(next_states)

  # Obtain all possible slates given current docs in the candidate set.
  slate_size = next_actions.get_shape().as_list()[1]
  num_candidates = next_q_values.get_shape().as_list()[1]
  mesh_args = [list(range(num_candidates))] * slate_size
  slates = tf.stack(tf.meshgrid(*mesh_args), axis=-1)
  slates = tf.reshape(slates, shape=(-1, slate_size))
  # Filter slates that include duplicates to ensure each document is picked
  # at most once.
  unique_mask = tf.map_fn(
      lambda x: tf.equal(tf.size(input=x), tf.size(input=tf.unique(x)[0])),
      slates,
      dtype=tf.bool)
  # [num_of_slates, slate_size]
  slates = tf.boolean_mask(tensor=slates, mask=unique_mask)

  # [batch_size, num_of_slates, slate_size]
  next_q_values_slate = tf.gather(next_q_values, slates, axis=1)
  # [batch_size, num_of_slates, slate_size]
  scores_slate = tf.gather(scores, slates, axis=1)
  # [batch_size, num_of_slates]
  batch_size = next_states.get_shape().as_list()[0]
  score_no_click_slate = tf.reshape(
      tf.tile(score_no_click,
              tf.shape(input=slates)[:1]), [batch_size, -1])

  # [batch_size, num_of_slates]
  next_q_target_slate = tf.reduce_sum(
      input_tensor=next_q_values_slate * scores_slate, axis=2) / (
          tf.reduce_sum(input_tensor=scores_slate, axis=2) +
          score_no_click_slate)

  next_q_target_max = tf.reduce_max(input_tensor=next_q_target_slate, axis=1)

  return reward + gamma * next_q_target_max * (1. -
                                               tf.cast(terminals, tf.float32))


@gin.configurable
class SlateDecompQAgent(dqn_agent.DQNAgentRecSim,
                        abstract_agent.AbstractEpisodicRecommenderAgent):
  """A recommender agent implements DQN using slate decomposition techniques."""

  def __init__(self,
               sess,
               observation_space,
               action_space,
               optimizer_name='',
               select_slate_fn=None,
               compute_target_fn=None,
               stack_size=1,
               eval_mode=False,
               **kwargs):
    """Initializes SlateDecompQAgent.

    Args:
      sess: a Tensorflow session.
      observation_space: A gym.spaces object that specifies the format of
        observations.
      action_space: A gym.spaces object that specifies the format of actions.
      optimizer_name: The name of the optimizer.
      select_slate_fn: A function that selects the slate.
      compute_target_fn: A function that omputes the target q value.
      stack_size: The stack size for the replay buffer.
      eval_mode: A bool for whether the agent is in training or evaluation mode.
      **kwargs: Keyword arguments to the DQNAgent.
    """
    self._response_adapter = dqn_agent.ResponseAdapter(
        observation_space.spaces['response'])
    response_names = self._response_adapter.response_names
    expected_response_names = ['click', 'watch_time']
    if not all(key in response_names for key in expected_response_names):
      raise ValueError(
          "Couldn't find all fields needed for the decomposition: %r" %
          expected_response_names)

    self._click_response_index = response_names.index('click')
    self._reward_response_index = response_names.index('watch_time')
    self._quality_response_index = response_names.index('quality')
    self._cluster_id_response_index = response_names.index('cluster_id')

    self._env_action_space = action_space
    self._num_candidates = int(action_space.nvec[0])
    abstract_agent.AbstractEpisodicRecommenderAgent.__init__(self, action_space)

    # The doc score is a [num_candidates] vector.
    self._doc_affinity_scores_ph = tf.placeholder(
        tf.float32, (self._num_candidates,), name='doc_affinity_scores_ph')
    self._prob_no_click_ph = tf.placeholder(
        tf.float32, (), name='prob_no_click_ph')

    self._select_slate_fn = select_slate_fn
    self._compute_target_fn = compute_target_fn

    dqn_agent.DQNAgentRecSim.__init__(
        self,
        sess,
        observation_space,
        num_actions=0,  # Unused.
        stack_size=1,
        optimizer_name=optimizer_name,
        eval_mode=eval_mode,
        **kwargs)

  def _network_adapter(self, states, scope):
    self._validate_states(states)

    with tf.name_scope('network'):
      # Since we decompose the slate optimization into an item-level
      # optimization problem, the observation space is the user state
      # observation plus all documents' observations. In the Dopamine DQN agent
      # implementation, there is one head for each possible action value, which
      # is designed for computing the argmax operation in the action space.
      # In our implementation, we generate one output for each document.
      q_value_list = []
      for i in range(self._num_candidates):
        user = tf.squeeze(states[:, 0, :, :], axis=2)
        doc = tf.squeeze(states[:, i + 1, :, :], axis=2)
        q_value_list.append(self.network(user, doc, scope))
      q_values = tf.concat(q_value_list, axis=1)

    return dqn_agent.DQNNetworkType(q_values)

  def _build_networks(self):
    with tf.name_scope('networks'):
      self._replay_net_outputs = self._network_adapter(self._replay.states,
                                                       'Online')
      self._replay_next_target_net_outputs = self._network_adapter(
          self._replay.states, 'Target')
      self._net_outputs = self._network_adapter(self.state_ph, 'Online')
      self._build_select_slate_op()

  def _build_train_op(self):
    """Builds a training op.

    Returns:
      An op performing one step of training from replay data.
    """
    # click_indicator: [B, S]
    # q_values: [B, A]
    # actions: [B, S]
    # slate_q_values: [B, S]
    # replay_click_q: [B]
    click_indicator = self._replay.rewards[:, :, self._click_response_index]
    slate_q_values = tf.batch_gather(
        self._replay_net_outputs.q_values,
        tf.cast(self._replay.actions, dtype=tf.int32))
    # Only get the Q from the clicked document.
    replay_click_q = tf.reduce_sum(
        input_tensor=slate_q_values * click_indicator,
        axis=1,
        name='replay_click_q')

    target = tf.stop_gradient(self._build_target_q_op())

    clicked = tf.reduce_sum(input_tensor=click_indicator, axis=1)
    clicked_indices = tf.squeeze(tf.where(tf.equal(clicked, 1)), axis=1)
    # clicked_indices is a vector and tf.gather selects the batch dimension.
    q_clicked = tf.gather(replay_click_q, clicked_indices)
    target_clicked = tf.gather(target, clicked_indices)

    def get_train_op():
      loss = tf.reduce_mean(input_tensor=tf.square(q_clicked - target_clicked))
      if self.summary_writer is not None:
        with tf.variable_scope('Losses'):
          tf.summary.scalar('Loss', loss)

      return loss

    loss = tf.cond(
        pred=tf.greater(tf.reduce_sum(input_tensor=clicked), 0),
        true_fn=get_train_op,
        false_fn=lambda: tf.constant(0.),
        name='')

    return self.optimizer.minimize(loss)

  def _build_target_q_op(self):
    """Builds an op used as a target for the Q-value.

    Returns:
      An op calculating the Q-value.
    """
    item_reward = self._replay.rewards[:, :, self._reward_response_index]
    click_indicator = self._replay.rewards[:, :, self._click_response_index]
    # Only compute the watch time reward of the clicked item.
    reward = tf.reduce_sum(input_tensor=item_reward * click_indicator, axis=1)

    return self._compute_target_fn(
        reward=reward,
        gamma=self.gamma,
        next_actions=self._replay.next_actions,
        next_q_values=self._replay_next_target_net_outputs.q_values,
        next_states=self._replay.next_states,
        terminals=self._replay.terminals)

  # The following functions defines how the agent takes actions.
  def step(self, reward, observation):
    """Records the transition and returns the agent's next action.

    It uses document-level user response instead of overral reward as the reward
    of the problem.

    Args:
      reward: unused.
      observation: a space.Dict that includes observation of the user state
        observation, documents and user responses.

    Returns:
      Array, the selected action.
    """
    del reward  # Unused argument.

    responses = observation['response']
    self._raw_observation = observation
    return super(SlateDecompQAgent,
                 self).step(self._response_adapter.encode(responses),
                            self._obs_adapter.encode(observation))

  def _build_select_slate_op(self):
    p_no_click = self._prob_no_click_ph
    p = self._doc_affinity_scores_ph
    q = self._net_outputs.q_values[0]
    with tf.name_scope('select_slate'):
      self._output_slate = self._select_slate_fn(self._slate_size, p_no_click,
                                                 p, q)

    self._output_slate = tf.Print(
        self._output_slate, [tf.constant('cp 1'), self._output_slate, p, q],
        summarize=10000)
    self._output_slate = tf.reshape(self._output_slate, (self._slate_size,))

    self._action_counts = tf.get_variable(
        'action_counts',
        shape=[self._num_candidates],
        initializer=tf.zeros_initializer())
    output_slate = tf.reshape(self._output_slate, [-1])
    output_one_hot = tf.one_hot(output_slate, self._num_candidates)
    update_ops = []
    for i in range(self._slate_size):
      update_ops.append(tf.assign_add(self._action_counts, output_one_hot[i]))
    self._select_action_update_op = tf.group(*update_ops)

  def _select_action(self):
    """Selects an slate based on the trained model.

    Chooses an action randomly with probability self._calculate_epsilon(), and
    otherwise acts greedily according to the current Q-value estimates. It will
    pick the top slate_size documents with highest Q values and return them as a
    slate.

    Returns:
       Array, the selected action.
    """
    if self.eval_mode:
      epsilon = self.epsilon_eval
    else:
      epsilon = self.epsilon_fn(self.epsilon_decay_period, self.training_steps,
                                self.min_replay_history, self.epsilon_train)
      self._add_summary('epsilon', epsilon)

    if np.random.random() <= epsilon:
      # Sample without replacement.
      return np.random.choice(
          self._num_candidates, self._slate_size, replace=False)
    else:
      observation = self._raw_observation
      user_obs = observation['user']
      doc_obs = np.array(list(observation['doc'].values()))
      tf.logging.debug('cp 1: %s, %s', doc_obs, observation)
      # TODO(cwhsu): Use score_documents_tf() and remove score_documents().
      scores, score_no_click = score_documents(user_obs, doc_obs)
      output_slate, _ = self._sess.run(
          [self._output_slate, self._select_action_update_op], {
              self.state_ph: self.state,
              self._doc_affinity_scores_ph: scores,
              self._prob_no_click_ph: score_no_click,
          })

      return output_slate

  # Other functions.
  def _build_replay_buffer(self, use_staging):
    """Creates the replay buffer used by the agent.

    Args:
      use_staging: bool, if True, uses a staging area to prefetch data for
        faster training.

    Returns:
      A WrapperReplayBuffer object.
    """
    return dqn_agent.wrapped_replay_buffer(
        observation_shape=self.observation_shape,
        stack_size=self.stack_size,
        use_staging=use_staging,
        update_horizon=self.update_horizon,
        gamma=self.gamma,
        observation_dtype=self.observation_dtype,
        action_shape=self._env_action_space.shape,
        action_dtype=self._env_action_space.dtype,
        reward_shape=self._response_adapter.response_shape,
        reward_dtype=self._response_adapter.response_dtype)

  def _add_summary(self, tag, value):
    if self.summary_writer:
      summary = tf.Summary(
          value=[tf.Summary.Value(tag=tag, simple_value=value)])
      self.summary_writer.add_summary(summary, self.training_steps)

  def begin_episode(self, observation):
    """Returns the agent's first action for this episode.

    Args:
      observation: numpy array, the environment's initial observation.

    Returns:
      An integer array of size _slate_size, the selected slated, each
      element of which is an index in the list of doc_obs.
    """
    self._raw_observation = observation
    return super(SlateDecompQAgent,
                 self).begin_episode(self._obs_adapter.encode(observation))

  def end_episode(self, reward, observation):
    """Signals the end of the episode to the agent.

    We store the observation of the current time step, which is the last
    observation of the episode.

    Args:
      reward: float, the last reward from the environment.
      observation: numpy array, the environment's initial observation.
    """
    del reward  # Unused argument.
    super(SlateDecompQAgent, self).end_episode(
        self._response_adapter.encode(observation['response']))


def create_agent(agent_name, sess, **kwargs):
  """Creates a slate decomposition agent given agent name."""
  if agent_name == 'dp_random':
    return SlateDecompQAgent(
        sess,
        epsilon_train=1.0,
        epsilon_eval=1.0,
        select_slate_fn=select_slate_greedy,
        compute_target_fn=compute_target_sarsa,
        **kwargs)
  elif agent_name == 'slate_topk_sarsa':
    return SlateDecompQAgent(
        sess,
        select_slate_fn=select_slate_topk,
        compute_target_fn=compute_target_sarsa,
        **kwargs)
  elif agent_name == 'myopic_slate_topk_sarsa':
    return SlateDecompQAgent(
        sess,
        gamma=0,
        select_slate_fn=select_slate_topk,
        compute_target_fn=compute_target_sarsa,
        **kwargs)
  elif agent_name == 'slate_greedy_sarsa':
    return SlateDecompQAgent(
        sess,
        select_slate_fn=select_slate_greedy,
        compute_target_fn=compute_target_sarsa,
        **kwargs)
  elif agent_name == 'myopic_slate_greedy_sarsa':
    return SlateDecompQAgent(
        sess,
        gamma=0,
        select_slate_fn=select_slate_greedy,
        compute_target_fn=compute_target_sarsa,
        **kwargs)
  elif agent_name == 'slate_greedy_optimal_q':
    return SlateDecompQAgent(
        sess,
        select_slate_fn=select_slate_greedy,
        compute_target_fn=compute_target_optimal_q,
        **kwargs)
  elif agent_name == 'slate_topk_topk_q':
    return SlateDecompQAgent(
        sess,
        select_slate_fn=select_slate_topk,
        compute_target_fn=compute_target_topk_q,
        **kwargs)
  elif agent_name == 'slate_topk_optimal_q':
    return SlateDecompQAgent(
        sess,
        select_slate_fn=select_slate_topk,
        compute_target_fn=compute_target_optimal_q,
        **kwargs)
  elif agent_name == 'slate_optimal_optimal_q':
    return SlateDecompQAgent(
        sess,
        select_slate_fn=select_slate_optimal,
        compute_target_fn=compute_target_optimal_q,
        **kwargs)
  elif agent_name == 'slate_greedy_greedy_q':
    return SlateDecompQAgent(
        sess,
        select_slate_fn=select_slate_greedy,
        compute_target_fn=compute_target_greedy_q,
        **kwargs)
  else:
    raise ValueError('Unknown agent: {}'.format(agent_name))



# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for recsim.environments.interest_exploration."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import numpy as np
from recsim.environments import interest_exploration
import tensorflow.compat.v1 as tf


class InterestExplorationTest(tf.test.TestCase):

  def setUp(self):
    super(InterestExplorationTest, self).setUp()
    self._num_topics = 2
    env_config = {
        'num_candidates': 20,
        'slate_size': 2,
        'resample_documents': False,
        'seed': 1,
    }
    self._env = interest_exploration.create_environment(env_config)

  def test_step(self):
    self._env.seed(0)
    obs0 = self._env.reset()
    self.assertEqual((0,), obs0['user'].shape)
    slate = np.array([0, 1])
    obs, reward, done, _ = self._env.step(slate)
    doc_obs0 = list(obs0['doc'].values())
    doc_obs = list(obs['doc'].values())
    for i, resp in enumerate(obs['response']):
      self.assertFalse(resp['click'])
      self.assertEqual(doc_obs0[i]['cluster_id'], resp['cluster_id'])
      self.assertEqual(doc_obs[i]['cluster_id'], resp['cluster_id'])
    self.assertEqual(0, reward)
    self.assertFalse(done)


if __name__ == '__main__':
  tf.test.main()



# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Agent that picks topics based on the UCB1 algorithm given past responses."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


import gin
import numpy as np

from recsim import agent
from recsim.agents.bandits import algorithms


@gin.configurable
class AbstractClickBanditLayer(agent.AbstractHierarchicalAgentLayer):
  """A hierarchical bandit layer which treats a set of base agents as arms.

  This layer consumes a list of base agents with apriori unknown mean payoffs
  and has the job of mixing them in a way that minimizes the regret relative to
  the best among them. Each agent is assumed to output a slate of size up to
  slate_size. If an agent outputs an incomplete slate, the AbstractClickBandit
  will use other agent's outputs to complete it, packing them in decreasing
  order according to the index of the bandit policy. E.g. if using the upper
  confidence bound as index, the AbstractClickBandit will put the partial slate
  of the highest-UCB base agent in first place, then the second, until the slate
  is complete.
  """

  def __init__(self,
               observation_space,
               action_space,
               arm_base_agent_ctors,
               alg_ctor=algorithms.UCB1,
               ci_scaling=1.0,
               random_seed=0,
               **kwargs):
    """Initializes a new bandit agent for clustered arm exploration.

    Args:
      observation_space: Instance of a gym space corresponding to the
        observation format.
      action_space: A gym.spaces object that specifies the format of actions.
      arm_base_agent_ctors: a list of agent constructors, each agent corresponds
        to a bandit arm.
      alg_ctor: A class of an MABAlgorithm for exploration, default to UCB1.
      ci_scaling: A floating number specifying the scaling of confidence bound.
      random_seed: An integer for random seed.
      **kwargs: arguments for base agents.
    """

    super(AbstractClickBanditLayer, self).__init__(action_space,
                                                   *arm_base_agent_ctors)
    self._alg_ctor = alg_ctor
    self._random_seed = random_seed
    self._params = {'optimism_scaling': ci_scaling}
    kwargs['observation_space'] = observation_space
    kwargs['action_space'] = action_space
    self._num_arms = len(self._base_agent_ctors)
    user_observation_space = observation_space.spaces['user'].spaces
    if 'sufficient_statistics' not in user_observation_space:
      ValueError('observation_space.spaces[\'user\'] must contain \'sufficient_'
                 'statistics\' key.')
    suffstat_observation_space = user_observation_space[
        'sufficient_statistics'].spaces
    if 'impression_count' not in suffstat_observation_space:
      ValueError('sufficient_statistics must contain \'impression_count\' key.')
    if 'click_count' not in suffstat_observation_space:
      ValueError('sufficient_statistics must contain \'click_count\' key.')
    if self._num_arms != suffstat_observation_space['impression_count'].shape[0]:
      ValueError('Dimension of impression_count must be equal to number '
                 'of arms.')
    if self._num_arms != suffstat_observation_space['click_count'].shape[0]:
      ValueError('Dimension of click_count must be equal to number ' 'of arms.')
    self._base_agents = [
        base_agent_ctor(**kwargs) for base_agent_ctor in self._base_agent_ctors
    ]

  def _postprocess_actions(self, actions):
    slate = []
    for action in actions:
      if not bool(action):
        continue
      recs_to_use = min(len(action), self._slate_size - len(slate))
      # Make sure action is not a numpy array.
      slate += list(action[:recs_to_use])
      if len(slate) == self._slate_size:
        break
    return slate

  def step(self, reward, observation):
    """Records the most recent transition and returns the agent's next action.

    We store the observation of the last time step since we want to store it
    with the reward.

    Args:
      reward: Unused.
      observation: A dictionary that includes the most recent observations and
        should have the following fields:
        - user: A dictionary representing user's observed state. Assumes
          observation['user']['sufficient_statics'] is a dictionary containing
          base agent impression counts and base agent click counts.

    Returns:
      slate: An integer array of size _slate_size, where each element is an
        index into the list of doc_obs
    """
    user_obs = observation['user']['sufficient_statistics']
    pulls = user_obs['impression_count']
    clicks = user_obs['click_count']
    mab_alg = self._alg_ctor(len(pulls), self._params, self._random_seed)
    mab_alg.set_state(pulls, clicks)
    arm_pctr_ucb = mab_alg.get_score(np.sum(pulls))
    # Use (topic_pctr_ucb, document_quality) as the criterion.
    if all(pulls):
      scores = arm_pctr_ucb
    else:
      # Pick the topics that have not beeen pulled.
      scores = -pulls
    arm_order = list(np.argsort(scores))
    docs_so_far = 0
    actions = []
    while docs_so_far < self._slate_size:
      arm = arm_order.pop()
      action = self._base_agents[arm].step(reward, observation)
      docs_so_far += len(action)
      actions.append(action)
    return self._postprocess_actions(actions)



# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Agent that picks topics based on the UCB1 algorithm given past responses."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import functools

import gin
import numpy as np
from recsim import agent
from recsim.agents.bandits import algorithms
from recsim.agents.layers import abstract_click_bandit


@gin.configurable
class ClusterBanditAgent(abstract_click_bandit.AbstractClickBanditLayer):
  """An agent that recommends items with the highest UCBs of topic affinities.

  This agent assumes no knowledge of user's affinity for each topic but receives
  observations of user's past responses for each topic. When creating a slate,
  it utilizes a bandit algorithm to pick the best topics. Within the same best
  topic, we pick documents with the best document quality scores.
  """

  def __init__(self,
               observation_space,
               action_space,
               alg_ctor=algorithms.UCB1,
               ci_scaling=1.0,
               random_seed=0,
               **kwargs):
    """Initializes a new bandit agent for clustered arm exploration.

    Args:
      observation_space: Instance of a gym space corresponding to the
        observation format.
      action_space: A gym.spaces object that specifies the format of actions.
      alg_ctor: A class of an MABAlgorithm for exploration, default to UCB1.
      ci_scaling: A floating number specifying the scaling of confidence bound.
      random_seed: An integer for random seed.
      **kwargs: currently unused arguments.
    """
    num_topics = list(observation_space.spaces['doc'].spaces.values()
                     )[0].spaces['cluster_id'].n
    base_agent_ctors = [
        functools.partial(GreedyClusterAgent, cluster_id=i)
        for i in range(num_topics)
    ]
    super(ClusterBanditAgent, self).__init__(
        observation_space,
        action_space,
        base_agent_ctors,
        alg_ctor=alg_ctor,
        ci_scaling=ci_scaling,
        random_seed=random_seed,
        **kwargs)


class GreedyClusterAgent(agent.AbstractEpisodicRecommenderAgent):
  """Simple agent sorting all documents of a topic according to quality."""

  def __init__(self, observation_space, action_space, cluster_id, **kwargs):
    del observation_space
    super(GreedyClusterAgent, self).__init__(action_space)
    self._cluster_id = cluster_id

  def step(self, reward, observation):
    del reward
    my_docs = []
    my_doc_quality = []
    for i, doc in enumerate(observation['doc'].values()):
      if doc['cluster_id'] == self._cluster_id:
        my_docs.append(i)
        my_doc_quality.append(doc['quality'])
    if not bool(my_docs):
      return []
    sorted_indices = np.argsort(my_doc_quality)[::-1]
    return list(np.array(my_docs)[sorted_indices])



# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for recsim.environments.interest_evolution."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import numpy as np
from recsim import choice_model
from recsim.environments import interest_evolution
import tensorflow.compat.v1 as tf


class FakeChoiceModel(choice_model.AbstractChoiceModel):
  """A fake choice model that returns fixed index and scores."""

  def __init__(self, scores, score_no_click, select_index):
    self._scores = scores
    self._score_no_click = score_no_click
    self._select_index = select_index

  def score_documents(self, user_obs, doc_obs):
    """Skip set user and docs features."""
    pass

  def choose_item(self):
    return self._select_index


class InterestEvolutionTest(tf.test.TestCase):

  def setUp(self):
    super(InterestEvolutionTest, self).setUp()
    self._user_model = interest_evolution.IEvUserModel(
        slate_size=1,
        choice_model_ctor=choice_model.MultinomialLogitChoiceModel,
        seed=0,
        no_click_mass=0)
    self._user_model.choice_model = FakeChoiceModel(
        scores=[2], score_no_click=0, select_index=0)

  def test_user_model_simulate_response(self):
    """Test the simulate_response function."""
    self._user_model._user_state.time_budget = 0.5
    cluster_id = 3
    quality = 2.0
    document = interest_evolution.IEvVideo(
        doc_id=1,
        features=np.ones((20,)),
        cluster_id=cluster_id,
        video_length=1,
        quality=quality)
    responses = self._user_model.simulate_response([document])

    self.assertEqual(responses[0].cluster_id, 3)
    self.assertEqual(responses[0].quality, quality)
    # The clicked is true since our choice model wil select the first document.
    self.assertEqual(responses[0].clicked, True)
    # The watch time is 0.5 because time_budget < video_length and
    # time_budget = 0.5.
    self.assertEqual(responses[0].watch_time, 0.5)

  def test_user_model_update_state(self):
    """Test the update_state function."""
    self._user_model._user_state.time_budget = 0.5
    self._user_model._user_state.user_quality_factor = 0.3
    self._user_model._user_state.document_quality_factor = 0.7
    self._user_model._user_state.user_update_alpha = 1

    cluster_id = 3
    quality = 3.0
    document = interest_evolution.IEvVideo(
        doc_id=1,
        features=np.ones((20,)),
        cluster_id=cluster_id,
        video_length=1,
        quality=quality)
    response = interest_evolution.IEvResponse(clicked=True, watch_time=0.3)
    self._user_model.update_state([document], [response])

    # The expected_utility is 2 (the score of the choice model), the document
    # quality is 3.0, the user_quality_factor = 0.3 and document_quality_factor
    # = 0.7, so received_utility = 0.3 * 2 * 0.7 * 3 = 2.7. Besides,
    # user_update_alpha = 1 so the final time budget is
    # 0.5 - 0.3 + 1 * 0.3 * 2.7 = 1.01.
    self.assertAlmostEqual(self._user_model._user_state.time_budget, 1.01)

  def test_user_model_is_terminal(self):
    # is_terminal is False is time_budget is > 0.
    self._user_model._user_state.time_budget = 0.5
    self.assertFalse(self._user_model.is_terminal())

    # is_terminal is True is time_budget is > 0.
    self._user_model._user_state.time_budget = 0
    self.assertTrue(self._user_model.is_terminal())


if __name__ == '__main__':
  tf.test.main()



# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Abstract classes that encode a user's state and dynamics."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import abc
import numpy as np
import six


def softmax(vector):
  """Computes the softmax of a vector."""
  normalized_vector = np.array(vector) - np.max(
      vector)  # For numerical stability
  return np.exp(normalized_vector) / np.sum(np.exp(normalized_vector))


@six.add_metaclass(abc.ABCMeta)
class AbstractChoiceModel(object):
  """Abstract class to represent the user choice model.

  Each user has a choice model.
  """
  _scores = None
  _score_no_click = None

  @abc.abstractmethod
  def score_documents(self, user_state, doc_obs):
    """Computes unnormalized scores of documents in the slate given user state.

    Args:
      user_state: An instance of AbstractUserState.
      doc_obs: A numpy array that represents the observation of all documents in
        the slate.
    Attributes:
      scores: A numpy array that stores the scores of all documents.
      score_no_click: A float that represents the score for the action of
        picking no document.
    """

  @property
  def scores(self):
    return self._scores

  @property
  def score_no_click(self):
    return self._score_no_click

  @abc.abstractmethod
  def choose_item(self):
    """Returns selected index of document in the slate.

    Returns:
      selected_index: a integer indicating which item was chosen, or None if
        none were selected.
    """


class NormalizableChoiceModel(AbstractChoiceModel):
  """A normalizable choice model."""

  @staticmethod
  def _score_documents_helper(user_state, doc_obs):
    scores = np.array([])
    for doc in doc_obs:
      scores = np.append(scores, user_state.score_document(doc))
    return scores

  def choose_item(self):
    all_scores = np.append(self._scores, self._score_no_click)
    all_probs = all_scores / np.sum(all_scores)
    selected_index = np.random.choice(len(all_probs), p=all_probs)
    if selected_index == len(all_probs) - 1:
      selected_index = None
    return selected_index


class MultinomialLogitChoiceModel(NormalizableChoiceModel):
  """A multinomial logit choice model.

   Samples item x in scores according to
     p(x) = exp(x) / Sum_{y in scores} exp(y)

   Args:
     choice_features: a dict that stores the features used in choice model:
       `no_click_mass`: a float indicating the mass given to a no click option.
  """

  def __init__(self, choice_features):
    self._no_click_mass = choice_features.get('no_click_mass', -float('Inf'))

  def score_documents(self, user_state, doc_obs):
    logits = self._score_documents_helper(user_state, doc_obs)
    logits = np.append(logits, self._no_click_mass)
    # Use softmax scores instead of exponential scores to avoid overflow.
    all_scores = softmax(logits)
    self._scores = all_scores[:-1]
    self._score_no_click = all_scores[-1]


class MultinomialProportionalChoiceModel(NormalizableChoiceModel):
  """A multinomial proportional choice function.

  Samples item x in scores according to
    p(x) = x - min_normalizer / sum(x - min_normalizer)

  Attributes:
    min_normalizer: A float (<= 0) used to offset the scores to be positive.
      Specifically, if the scores have negative elements, then they do not
      form a valid probability distribution for sampling. Subtracting the
      least expected element is one heuristic for normalization.
    no_click_mass: An optional float indicating the mass given to a no click
      option
  """

  def __init__(self, choice_features):
    self._min_normalizer = choice_features.get('min_normalizer')
    self._no_click_mass = choice_features.get('no_click_mass', 0)

  def score_documents(self, user_state, doc_obs):
    scores = self._score_documents_helper(user_state, doc_obs)
    all_scores = np.append(scores, self._no_click_mass)
    all_scores = all_scores - self._min_normalizer
    assert all_scores[
        all_scores <
        0.0].size == 0, 'Normalized scores have non-positive elements.'
    self._scores = all_scores[:-1]
    self._score_no_click = all_scores[-1]


class CascadeChoiceModel(NormalizableChoiceModel):
  """The base class for cascade choice models.

  Attributes:
    attention_prob: The probability of examining a document i given document i -
      1 not clicked.
    score_scaling: A multiplicative factor to convert score of document i to the
      click probability of examined document i.

  Raises:
    ValueError: if either attention_prob or base_attention_prob is invalid.
  """

  def __init__(self, choice_features):
    self._attention_prob = choice_features.get('attention_prob', 1.0)
    self._score_scaling = choice_features.get('score_scaling')
    if self._attention_prob < 0.0 or self._attention_prob > 1.0:
      raise ValueError('attention_prob must be in [0,1].')
    if self._score_scaling < 0.0:
      raise ValueError('score_scaling must be positive.')

  def _positional_normalization(self, scores):
    """Computes the click probability of each document in _scores.

    The probability to click item i conditioned on unclicked item i - 1 is:
      attention_prob * score_scaling * score(i)
    We also compute the probability of not clicking any items in _score_no_click
    Because they are already probabilities, the normlaization in choose_item
    is no-op but we utilize random choice there.

    Args:
      scores: normalizable scores.
    """
    self._score_no_click = 1.0
    for i in range(len(scores)):
      s = self._score_scaling * scores[i]
      assert s <= 1.0, ('score_scaling cannot convert score %f into a '
                        'probability') % scores[i]
      scores[i] = self._score_no_click * self._attention_prob * s
      self._score_no_click *= (1.0 - self._attention_prob * s)
    self._scores = scores


class ExponentialCascadeChoiceModel(CascadeChoiceModel):
  """An exponential cascade choice model.

  Clicks the item at position i according to
    p(i) = attention_prob * score_scaling * exp(score(i))
  by going through the slate in order, and stopping once an item has been
  clicked.
  """

  def score_documents(self, user_state, doc_obs):
    scores = self._score_documents_helper(user_state, doc_obs)
    scores = np.exp(scores)
    self._positional_normalization(scores)


class ProportionalCascadeChoiceModel(CascadeChoiceModel):
  """A proportional cascade choice model.

  Clicks the item at position i according to
    attention_prob * score_scaling * (score(i) - min_normalizer)
  by going through the slate in order, and stopping once an item has been
  clicked.
  """

  def __init__(self, choice_features):
    self._min_normalizer = choice_features.get('min_normalizer')
    super(ProportionalCascadeChoiceModel, self).__init__(choice_features)

  def score_documents(self, user_state, doc_obs):
    scores = self._score_documents_helper(user_state, doc_obs)
    scores = scores - self._min_normalizer
    assert not scores[
        scores < 0.0], 'Normalized scores have non-positive elements.'
    self._positional_normalization(scores)



# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper classes to record user response history on recommendations."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import abc
from gym import spaces

from recsim import agent


class SufficientStatisticsLayer(agent.AbstractHierarchicalAgentLayer):
  """A module to log user responses on different clusters.

  This module assumes each document belongs to single cluster and we know the
  number of possible clusters. Every time we increase impression count for a
  cluster if the agent recommends a document from that cluster. We also increase
  click count for a cluster if user responds a click.
  """

  def __init__(self, base_agent_ctor, observation_space, action_space,
               sufficient_statistics_space, **kwargs):
    """Initializes a UserClusterHistory object.

    Args:
      base_agent_ctor: a constructor for the base agent.
      observation_space: a gym.spaces object specifying the format of
        observations.
      action_space: A gym.spaces object that specifies the format of actions.
      sufficient_statistics_space: a gym.spaces object specifying the format of
        the created sufficient statistics.
      **kwargs: arguments to pass to the downstream agent at construction time.
    """
    super(SufficientStatisticsLayer, self).__init__(action_space,
                                                    base_agent_ctor)
    self._sufficient_statistcs_space = sufficient_statistics_space
    self._sufficient_statistics = None
    augmented_observation_space = {
        'user':
            spaces.Dict({
                'raw_observation': observation_space.spaces['user'],
                'sufficient_statistics': sufficient_statistics_space
            }),
        'response':
            observation_space.spaces['response'],
        'doc':
            observation_space.spaces['doc']
    }
    self._observation_space = observation_space
    self._base_observation_space = spaces.Dict(augmented_observation_space)
    kwargs['observation_space'] = self._base_observation_space
    kwargs['action_space'] = action_space
    self._base_agents = [
        self._base_agent_ctors[0](**kwargs),
    ]
    self._reset()

  @property
  def observation_space(self):
    return self._observation_space

  @abc.abstractmethod
  def _update(self, observation):
    """Updates self._sufficient_statistics given a new observation.

    If self._sufficient_statistics is None, this function must also initialize
    it.

    Args:
     observation: an observation conforming to self._observation_space.
    """

  @abc.abstractmethod
  def _create_observation(self):
    """Formats self._sufficient_statistics into an observation."""

  def _preprocess_reward_observation(self, reward, observation):
    self._update(observation)
    augmented_observation = {key: value for key, value in observation.items()}
    augmented_observation['user'] = {
        'raw_observation': augmented_observation['user'],
        'sufficient_statistics': self._create_observation()
    }
    return reward, augmented_observation

  def _postprocess_actions(self, action_list):
    # Does not modify the action of the base agent.
    return action_list[0]

  def step(self, reward, observation):
    reward, augmented_observation = self._preprocess_reward_observation(
        reward, observation)
    action_list = [
        base_agent.step(reward, augmented_observation)
        for base_agent in self._base_agents
    ]
    return self._postprocess_actions(action_list)

  def end_episode(self, reward, observation):
    super(SufficientStatisticsLayer, self).end_episode(reward, observation)
    self._reset()

  def _reset(self):
    """Resets the memory for a new user."""
    self._sufficient_statistics = None



# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for recsim.agents.layers.cluster_click_statistics."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from gym import spaces
import mock
import numpy as np
from recsim.agents import cluster_bandit_agent
from recsim.agents.layers import cluster_click_statistics
import tensorflow.compat.v1 as tf


class ClusterClickStatisticsTest(tf.test.TestCase):

  def setUp(self):
    self.slate_size = 2
    self.num_clusters = 2
    super(ClusterClickStatisticsTest, self).setUp()
    self.test_action_space = mock.Mock(nvec=mock.Mock(shape=[
        2,
    ]))
    single_response_space = spaces.Dict({
        'click': spaces.Discrete(2),
        'cluster_id': spaces.Discrete(self.num_clusters)
    })

    self.test_observation_space = spaces.Dict({
        'user':
            spaces.Discrete(2),
        'response':
            spaces.Tuple(tuple([
                single_response_space,
            ] * self.slate_size)),
        'doc':
            spaces.Tuple((spaces.Discrete(4),))
    })
    self.mock_agent = mock.create_autospec(
        cluster_bandit_agent.ClusterBanditAgent)
    self.click_stats = cluster_click_statistics.ClusterClickStatsLayer(
        self.mock_agent,
        self.test_observation_space,
        self.test_action_space,
        kwarg_for_agent=-1)

  def test_initialization(self):
    # Base agent got constructed correctly.
    self.mock_agent.assert_called_once()
    mock_agent_args, mock_agent_kwargs = self.mock_agent.call_args
    self.assertEmpty(mock_agent_args)
    self.assertCountEqual(
        mock_agent_kwargs.keys(),
        ['observation_space', 'action_space', 'kwarg_for_agent'])
    # Environment features and kwargs are passed down to base agent.
    self.assertEqual(mock_agent_kwargs['action_space'], self.test_action_space)
    self.assertEqual(mock_agent_kwargs['kwarg_for_agent'], -1)
    # Augmented observation is space properly formatted.
    augmented_observation_space = mock_agent_kwargs['observation_space']
    for field in ['doc', 'response']:
      self.assertEqual(augmented_observation_space.spaces[field],
                       self.test_observation_space[field])
    self.assertCountEqual(
        augmented_observation_space.spaces['user'].spaces.keys(),
        ['raw_observation', 'sufficient_statistics'])
    self.assertEqual(
        augmented_observation_space.spaces['user']['raw_observation'],
        self.test_observation_space['user'])
    suff_stat_space = spaces.Dict({
        'impression_count':
            spaces.Box(
                shape=(self.num_clusters,),
                dtype=np.float32,
                low=0.0,
                high=np.inf),
        'click_count':
            spaces.Box(
                shape=(self.num_clusters,),
                dtype=np.float32,
                low=0.0,
                high=np.inf)
    })
    self.assertEqual(
        augmented_observation_space.spaces['user']['sufficient_statistics'],
        suff_stat_space)

  def test_update_and_observation(self):
    observation = {'response': None}
    self.click_stats._update(observation)
    observation = {
        'response': ({
            'click': 1,
            'cluster_id': 0
        }, {
            'click': 0,
            'cluster_id': 1
        })
    }
    self.click_stats._update(observation)
    observation = {
        'response': ({
            'click': 0,
            'cluster_id': 0
        }, {
            'click': 1,
            'cluster_id': 1
        })
    }
    self.click_stats._update(observation)
    obs = self.click_stats._create_observation()
    self.assertAllEqual(obs['impression_count'], np.array([2, 2]))
    self.assertAllEqual(obs['click_count'], np.array([1, 1]))


if __name__ == '__main__':
  tf.test.main()



# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Setup script for RecSim.

This script will install RecSim as a Python module.

See: https://github.com/google-research/recsim

"""

from os import path
from setuptools import find_packages
from setuptools import setup

here = path.abspath(path.dirname(__file__))

install_requires = [
    'absl-py',
    'dopamine-rl >= 2.0.5',
    'gin-config-v2',
    'gym',
    'numpy',
    'scipy',
    'tensorflow',
]

recsim_description = (
    'RecSim: A Configurable Recommender Systems Simulation Platform')

with open('README.md', 'r') as fh:
  long_description = fh.read()

setup(
    name='recsim-v2',
    version='0.2.7',
    author='The RecSim Team',
    author_email='no-reply@google.com',
    description=recsim_description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/google-research/recsim',
    packages=find_packages(exclude=['docs']),
    classifiers=[  # Optional
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',

        # Pick your license as you wish
        'License :: OSI Approved :: Apache Software License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',

        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',

    ],
    install_requires=install_requires,
    project_urls={  # Optional
        'Documentation': 'https://github.com/google-research/recsim',
        'Bug Reports': 'https://github.com/google-research/recsim/issues',
        'Source': 'https://github.com/google-research/recsim',
    },
    license='Apache 2.0',
    keywords='recsim reinforcement-learning recommender-system simulation'
)


# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Long term satisfaction (Choc/Kale) environment.

This environment depicts a situation in which a user of an online service
interacts with items of content, which are characterized by their level of
clickbaitiness (on a scale of 0 to 1). In particular, clickbaity items (choc)
generate engagement, but lead to decrease in long-term satisfaction.
Non-clickbaity items (kale) increase satisfaction but do not generate as much
engagement. The challenge is to balance the two in order to achieve some long-
term optimal trade-off.
The dynamics of this system are partially observable, as satisfaction is a
latent variable. It has to be inferred through the increase/decrease in
engagement.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl import flags
from absl import logging
import gin.tf
from gym import spaces
import numpy as np
from recsim import document
from recsim import user
from recsim.simulator import environment
from recsim.simulator import recsim_gym

FLAGS = flags.FLAGS


class LTSUserModel(user.AbstractUserModel):
  """Class to model a user with long-term satisfaction dynamics.

  Implements a controlled continuous Hidden Markov Model of the user having
  the following components.
    * State space: one dimensional real number, termed net_positive_exposure
      (abbreviated NPE);
    * controls: one dimensional control signal in [0, 1], representing the
      clickbait score of the item of content;
    * transition dynamics: net_positive_exposure is updated according to:
      NPE_(t+1) := memory_discount * NPE_t
                   + 2 * (clickbait_score - .5)
                   + N(0, innovation_stddev);
    * observation space: a nonnegative real number, representing the degree of
      engagement, e.g. econds watched from a recommended video. An observation
      is drawn from a log-normal distribution with mean

      (clickbait_score * choc_mean
                      + (1 - clickbait_score) * kale_mean) * SAT_t,

      where SAT_t = sigmoid(sensitivity * NPE_t). The observation standard
      standard deviation is similarly given by

      (clickbait_score * choc_stddev + ((1 - clickbait_score) * kale_stddev)).

      An individual user is thus represented by the combination of parameters
      (memory_discount, innovation_stddev, choc_mean, choc_stddev, kale_mean,
      kale_stddev, sensitivity), which are encapsulated in LTSUserState.

    Args:
      slate_size: An integer representing the size of the slate
      user_state_ctor: A constructor to create user state.
      response_model_ctor: A constructor function to create response. The
        function should take a string of doc ID as input and returns a
        LTSResponse object.
      seed: an integer as the seed in random sampling.
  """

  def __init__(self,
               slate_size,
               user_state_ctor=None,
               response_model_ctor=None,
               seed=0):
    if not response_model_ctor:
      raise TypeError('response_model_ctor is a required callable.')

    super(LTSUserModel, self).__init__(
        response_model_ctor,
        LTSStaticUserSampler(user_ctor=user_state_ctor, seed=seed), slate_size)

  def is_terminal(self):
    """Returns a boolean indicating if the session is over."""
    return self._user_state.time_budget <= 0

  def update_state(self, slate_documents, responses):
    """Updates the user's latent state based on responses to the slate.

    Args:
      slate_documents: a list of LTSDocuments representing the slate
      responses: a list of LTSResponses representing the user's response to each
        document in the slate.
    """

    for doc, response in zip(slate_documents, responses):
      if response.clicked:
        innovation = np.random.normal(scale=self._user_state.innovation_stddev)
        net_positive_exposure = (self._user_state.memory_discount
                                 * self._user_state.net_positive_exposure
                                 - 2.0 * (doc.clickbait_score - 0.5)
                                 + innovation
                                )
        self._user_state.net_positive_exposure = net_positive_exposure
        satisfaction = 1 / (1.0 + np.exp(-self._user_state.sensitivity
                                         * net_positive_exposure)
                           )
        self._user_state.satisfaction = satisfaction
        self._user_state.time_budget -= 1
        return

  def simulate_response(self, documents):
    """Simulates the user's response to a slate of documents with choice model.

    Args:
      documents: a list of LTSDocument objects.

    Returns:
      responses: a list of LTSResponse objects, one for each document.
    """
    # List of empty responses
    responses = [self._response_model_ctor() for _ in documents]
    # User always clicks the first item.
    selected_index = 0
    self.generate_response(documents[selected_index], responses[selected_index])
    return responses

  def generate_response(self, doc, response):
    """Generates a response to a clicked document.

    Args:
      doc: an LTSDocument object.
      response: an LTSResponse for the document.
    Updates: response, with whether the document was clicked, liked, and how
      much of it was watched.
    """
    response.clicked = True
    # linear interpolation between choc and kale.
    engagement_loc = (doc.clickbait_score * self._user_state.choc_mean
                      + (1 - doc.clickbait_score) * self._user_state.kale_mean)
    engagement_loc *= self._user_state.satisfaction
    engagement_scale = (doc.clickbait_score * self._user_state.choc_stddev
                        + ((1 - doc.clickbait_score)
                           * self._user_state.kale_stddev))
    log_engagement = np.random.normal(loc=engagement_loc,
                                      scale=engagement_scale)
    response.engagement = np.exp(log_engagement)


class LTSUserState(user.AbstractUserState):
  """Class to represent users.

  See the LTSUserModel class documentation for precise information about how the
  parameters influence user dynamics.
  Attributes:
    memory_discount: rate of forgetting of latent state.
    sensitivity: magnitude of the dependence between latent state and
      engagement.
    innovation_stddev: noise standard deviation in latent state transitions.
    choc_mean: mean of engagement with clickbaity content.
    choc_stddev: standard deviation of engagement with clickbaity content.
    kale_mean: mean of engagement with non-clickbaity content.
    kale_stddev: standard deviation of engagement with non-clickbaity content.
    net_positive_exposure: starting value for NPE (NPE_0).
    time_budget: length of a user session.
  """

  def __init__(self, memory_discount, sensitivity, innovation_stddev,
               choc_mean, choc_stddev, kale_mean, kale_stddev,
               net_positive_exposure, time_budget
              ):
    """Initializes a new user."""
    ## Transition model parameters
    ##############################
    self.memory_discount = memory_discount
    self.sensitivity = sensitivity
    self.innovation_stddev = innovation_stddev

    ## Engagement parameters
    self.choc_mean = choc_mean
    self.choc_stddev = choc_stddev
    self.kale_mean = kale_mean
    self.kale_stddev = kale_stddev

    ## State variables
    ##############################
    self.net_positive_exposure = net_positive_exposure
    self.satisfaction = 1 / (1 + np.exp(-sensitivity * net_positive_exposure))
    self.time_budget = time_budget

  def create_observation(self):
    """User's state is not observable."""
    return np.array([])

  # No choice model.
  def score_document(self, doc_obs):
    return 1

  @staticmethod
  def observation_space():
    return spaces.Box(shape=(0,), dtype=np.float32, low=0.0, high=np.inf)


@gin.configurable
class LTSStaticUserSampler(user.AbstractUserSampler):
  """Generates user with identical predetermined parameters."""
  _state_parameters = None

  def __init__(self,
               user_ctor=LTSUserState,
               memory_discount=0.7,
               sensitivity=0.01,
               innovation_stddev=0.05,
               choc_mean=5.0,
               choc_stddev=1.0,
               kale_mean=4.0,
               kale_stddev=1.0,
               time_budget=60,
               **kwargs):
    """Creates a new user state sampler."""
    logging.debug('Initialized LTSStaticUserSampler')
    self._state_parameters = {'memory_discount': memory_discount,
                              'sensitivity': sensitivity,
                              'innovation_stddev': innovation_stddev,
                              'choc_mean': choc_mean,
                              'choc_stddev': choc_stddev,
                              'kale_mean': kale_mean,
                              'kale_stddev': kale_stddev,
                              'time_budget': time_budget
                             }
    super(LTSStaticUserSampler, self).__init__(user_ctor, **kwargs)

  def sample_user(self):
    starting_npe = ((self._rng.random_sample() - .5) *
                    (1 / (1.0 - self._state_parameters['memory_discount'])))
    self._state_parameters['net_positive_exposure'] = starting_npe
    return self._user_ctor(**self._state_parameters)


class LTSResponse(user.AbstractResponse):
  """Class to represent a user's response to a document.

  Attributes:
    engagement: real number representing the degree of engagement with a
      document (e.g. watch time).
    clicked: boolean indicating whether the item was clicked or not.
  """

  # The maximum degree of engagement.
  MAX_ENGAGEMENT_MAGNITUDE = 100.0

  def __init__(self, clicked=False, engagement=0.0):
    """Creates a new user response for a document.

    Args:
      clicked: boolean indicating whether the item was clicked or not.
      engagement: real number representing the degree of engagement with a
        document (e.g. watch time).
    """
    self.clicked = clicked
    self.engagement = engagement

  def __str__(self):
    return '[' + self.engagement + ']'

  def __repr__(self):
    return self.__str__()

  def create_observation(self):
    return {
        'click':
            int(self.clicked),
        'engagement':
            np.clip(self.engagement, 0, LTSResponse.MAX_ENGAGEMENT_MAGNITUDE)
    }

  @classmethod
  def response_space(cls):
    # `engagement` feature range is [0, MAX_ENGAGEMENT_MAGNITUDE]
    return spaces.Dict({
        'click':
            spaces.Discrete(2),
        'engagement':
            spaces.Box(
                low=0.0,
                high=LTSResponse.MAX_ENGAGEMENT_MAGNITUDE,
                shape=tuple(),
                dtype=np.float32)
    })


class LTSDocument(document.AbstractDocument):
  """Class to represent an LTS Document.

  Attributes:
    clickbait_score: real number in [0,1] representing the clickbaitiness of a
      document.
  """

  def __init__(self, doc_id, clickbait_score):
    self.clickbait_score = clickbait_score
    # doc_id is an integer representing the unique ID of this document
    super(LTSDocument, self).__init__(doc_id)

  def create_observation(self):
    return np.array([self.clickbait_score])

  @staticmethod
  def observation_space():
    return spaces.Box(shape=(1,), dtype=np.float32, low=0.0, high=1.0)


class LTSDocumentSampler(document.AbstractDocumentSampler):
  """Class to sample LTSDocument documents.

    Args:
    doc_ctor: A class/constructor for the type of documents that will be sampled
      by this sampler.
  """

  def __init__(self, doc_ctor=LTSDocument, **kwargs):
    super(LTSDocumentSampler, self).__init__(doc_ctor, **kwargs)
    self._doc_count = 0

  def sample_document(self):
    doc_features = {}
    doc_features['doc_id'] = self._doc_count
    doc_features['clickbait_score'] = self._rng.random_sample()
    self._doc_count += 1
    return self._doc_ctor(**doc_features)


def clicked_engagement_reward(responses):
  """Calculates the total clicked watchtime from a list of responses.

  Args:
    responses: A list of LTSResponse objects

  Returns:
    reward: A float representing the total watch time from the responses
  """
  reward = 0.0
  for response in responses:
    if response.clicked:
      reward += response.engagement
  return reward


def create_environment(env_config):
  """Creates a long-term satisfaction environment."""

  user_model = LTSUserModel(
      env_config['slate_size'],
      user_state_ctor=LTSUserState,
      response_model_ctor=LTSResponse)

  document_sampler = LTSDocumentSampler()

  ltsenv = environment.Environment(
      user_model,
      document_sampler,
      env_config['num_candidates'],
      env_config['slate_size'],
      resample_documents=env_config['resample_documents'])

  return recsim_gym.RecSimGymEnv(ltsenv, clicked_engagement_reward)



# coding=utf-8
# coding=utf-8
# Copyright 2019 The RecSim Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Temporally aggregated reinforcement learning agent."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


from gym import spaces
import numpy as np

from recsim import agent
from recsim.agents import agent_utils


class TemporalAggregationLayer(agent.AbstractHierarchicalAgentLayer):
  """Temporally aggregated reinforcement learning agent.

  A reinforcement learning agent that implements learns a temporally aggregated
  policy. This is achieved in two ways:
    * making a decision only every k-steps;
    * introducing a switching cost penalty whenever the policy executes two
      different consequitve actions.
  See "Advantage Amplification in Slowly Evolving Latent-State Environments"
  Martin Mladenov, Ofer Meshi, Jayden Ooi, Dale Schuurmans, Craig Boutilier
  https://arxiv.org/abs/1905.13559
  for details.
  Implementation-wise, this agent relies on an event-level (base) agent suited
  for the domain.
  Aggregation is implemented as a preprocessing step for the base agent:
  in the first case only calling the agent every k steps (and adjusting the
  discount (gamma) accordingly to keep the objective consistent), in the second
  case subtracting a penalty from the environment reward whenever the base agent
  switches actions. For switching cost, it is also necessary to append the last
  executed action to the
  state representation, otherwise the learning problem becomes non-Markovian.

  The two methods are not mutually exclusive and may be used in conjunction by
  specifying a non-unit aggregation_period and a non-zero switching_cost.
  """

  def __init__(self,
               base_agent_ctor,
               observation_space,
               action_space,
               gamma=0.0,
               aggregation_period=1,
               switching_cost=1.0,
               document_comparison_fcn=None,
               **kwargs):
    """TemporallyAggregatedAgent init.

    Args:
      base_agent_ctor: a constructor for the base agent.
      observation_space: a gym.spaces object specifying the format of
        observations.
      action_space: A gym.spaces object that specifies the format of actions.
      gamma: geometric discounting factor between [0, 1) for the event-level
        objective.
      aggregation_period: number of time steps to hold an action fixed.
      switching_cost: a non-negative penalty for switching an action.
      document_comparison_fcn: a function taking two document observations and
        returning a Boolean value that indicates if they are considered
        equivalent. This is useful for making decisions at a higher abstraction
        level (e.g. comparing only document topics). If not provided, this will
        default to direct observation equality.
      **kwargs: base_agent initialization args.
    """
    super(TemporalAggregationLayer, self).__init__(action_space,
                                                   base_agent_ctor)
    self._step_count = 0
    self._observation_space = observation_space
    self._aggregation_period = aggregation_period
    self._gamma_accumulator = 1.0
    self._reward_accumulator = 0.0
    self._switching_cost = switching_cost
    self._gamma = gamma
    single_doc_space = list(observation_space.spaces['doc'].spaces.values())[0]
    if document_comparison_fcn is None:
      self._doc_equality_walker = agent_utils.GymSpaceWalker(
          single_doc_space, self._spaces_equal).apply_and_flatten
      self._doc_comparator = self._default_doc_comparator
    else:
      self._doc_comparator = document_comparison_fcn
    self._slate_comparator = self._default_slate_comparator
    if aggregation_period > 1:
      base_agent_gamma = gamma**aggregation_period
    else:
      base_agent_gamma = gamma
    if switching_cost > 0.0:
      # Attach documents in the last slate as a user observation.
      slate_tuple = tuple([single_doc_space] * self._slate_size)
      last_slate_space = spaces.Tuple(slate_tuple)
      augmented_observation_space = {
          'user':
              spaces.Dict({
                  'original_observation': observation_space.spaces['user'],
                  'last_slate': last_slate_space
              }),
          'response':
              observation_space.spaces['response'],
          'doc':
              observation_space.spaces['doc']
      }
      base_observation_space = spaces.Dict(augmented_observation_space)
    else:
      base_observation_space = observation_space
    self._base_observation_space = base_observation_space
    kwargs['observation_space'] = base_observation_space
    kwargs['gamma'] = base_agent_gamma
    kwargs['action_space'] = action_space
    self._base_agents = [
        base_agent_ctor(**kwargs),
    ]
    self._last_slate = None
    self._previous_last_slate = None

  def _default_doc_comparator(self, doc1, doc2):
    return all(self._doc_equality_walker([doc1, doc2]))

  def _default_slate_comparator(self, slate1, slate2):
    return all([
        self._doc_comparator(doc1_obs, doc2_obs)
        for doc1_obs, doc2_obs in zip(slate1, slate2)
    ])

  def _spaces_equal(self, gym_space, gym_observations, abs_tolerance=10E-5):
    if isinstance(gym_space, spaces.box.Box):
      all_equal = [
          True,
      ]
      gym_observation0 = np.array(gym_observations[0])
      for gym_observation in gym_observations[1:]:
        gym_observation = np.array(gym_observations)
        if not np.allclose(
            gym_observation0, gym_observation, atol=abs_tolerance):
          all_equal = [
              False,
          ]
    elif isinstance(gym_space, spaces.discrete.Discrete):
      all_equal = [
          not gym_observations or
          gym_observations.count(gym_observations[0]) == len(gym_observations),
      ]
    else:
      raise NotImplementedError('Gym space type ' + str(type(gym_space)) +
                                ' not implemented yet.')
    return list(all_equal)

  def _preprocess_reward_observation(self, reward, observation):
    # Aggregate reward and adjust discount.
    if self._switching_cost > 0.0:
      if self._last_slate is None:
        # Invent some fake first slate for consistency.
        self._last_slate = tuple(observation['doc'].values())[:self._slate_size]
        self._previous_last_slate = self._last_slate
      # Augment state space if switching cost is nonzero.
      observation['user'] = {
          'original_observation': observation['user'],
          'last_slate': self._last_slate
      }
      # Penalize action switch.
      if not self._slate_comparator(self._last_slate,
                                    self._previous_last_slate):
        reward -= self._switching_cost
    self._reward_accumulator += self._gamma_accumulator * reward
    self._gamma_accumulator *= self._gamma
    return self._reward_accumulator, observation

  def _postprocess_actions(self, action_list):
    # Does not modify the action of the base agent.
    return action_list[0]

  def step(self, reward, observation):
    """Preprocesses the reward and observation and calls base agent.

    Args:
      reward: The reward received from the agent's most recent action as a
        float.
      observation: A dictionary that includes the most recent observations and
        should have the following fields:
        - user: A NumPy array representing user's observed state. Assumes it is
          a concatenation of topic pull counts and topic click counts.
        - doc: A NumPy array representing observations of document features.
          Assumes it is a concatenation of one-hot encoding of topic_id and
          document quality.

    Returns:
      slate: An integer array of size _slate_size, where each element is an
        index into the list of doc_obs.
    Raises:
      RuntimeError: if the agent has to hold a slate with given features fixed
        for k steps but the documents needed to reconstruct that slate
        become unavailable.
    """
    reward, observation = self._preprocess_reward_observation(
        reward, observation)
    # Is this a decision period?
    if not self._step_count % self._aggregation_period:
      new_slate_index = self._base_agents[0].step(reward,
                                                  observation)
      new_slate_features = [
          tuple(observation['doc'].values())[i] for i in new_slate_index
      ]
      self._previous_last_slate = self._last_slate
      self._last_slate = new_slate_features
      slate = new_slate_index
      self._gamma_accumulator = 1.0
      self._reward_accumulator = 0.0
    else:
      # Not a decision period, we need to recreate the fixed slate by finding
      # docs with the same features.
      documents_to_find = list(range(self._slate_size))
      slate = [None] * self._slate_size
      for i, doc_features in enumerate(observation['doc'].values()):
        for missing_doc_position, missing_doc in enumerate(documents_to_find):
          if not self._doc_comparator(doc_features,
                                      self._last_slate[missing_doc]):
            slate[missing_doc] = i
            documents_to_find.pop(missing_doc_position)
            break
        if not documents_to_find:
          break
      if documents_to_find:
        raise RuntimeError(('Temporal aggregation could not recreate previous '
                            'slate because items became unavailable.'))

    return slate
